#ifndef CDDETECTOR_H
#define CDDETECTOR_H
#include "CDConvex.h"
#include <SceneGraph/SGFrame.h>
#include <Base/Combination.h>
#include "CDQuickHull3D.h"

namespace Spr {;

template <class T, class ITI, class ITJ, class GetContainer>
class UTComposedIterator{
public:
	typedef UTComposedIterator<T, ITI, ITJ, GetContainer> self_type;
protected:
	ITI iti, iEnd;
	ITJ itj;
public:
	self_type(){}
	T operator->(){ return (T&)*itj; }
	T operator*(){ return (T&)*itj; }
	self_type& operator ++(int){ Next(); return *this; }
	self_type operator ++(){ self_type rv(*this); Next(); return rv; }
	bool operator ==(const self_type& i) const{ return i.iti == iti && i.itj == itj; }
	bool operator !=(const self_type& i) const {return !(i==*this);}
	void Init(ITI b, ITI e, ITJ j){
		while (b!=e && !GetContainer()(b)) ++b;
		iti = b;
		iEnd = e;
		itj = j;
	}
	void InitAsBegin(ITI b, ITI e){
		Init(b, e, NULL);
		if (b!=e) itj = GetContainer()(iti)->begin();
		else return;
		while (itj==GetContainer()(iti)->end()){
			do{
				++iti;
				if (iti == iEnd){
					itj=NULL;
					break;
				}
			} while(!GetContainer()(iti));
			itj=GetContainer()(iti)->begin();
		}
	}
	void InitAsEnd(ITI b, ITI e){
		Init(e, e, 0);
	}
	void Next(){
		assert(itj != NULL);
		++itj;
		if (itj == GetContainer()(iti)->end()){
			do{
				++iti;
			}while(iti != iEnd && (!GetContainer()(iti) || !GetContainer()(iti)->size()));
			if (iti != iEnd){
				itj = GetContainer()(iti)->begin();
			}else{
				itj = NULL;
			}
		}
	}
};


///	ʌ`̃yA
class CDConvexPair:public UTRefCount{
public:
	virtual ~CDConvexPair(){}
	UTRef<CDConvex> convex[2];
	Vec3d closestPoint[2]; // [JWn
	Vec3d commonPoint;     // O[oWn
	unsigned int lastContactCount;
	enum State{
		NEW,
		CONTINUE,
	} state;
	void Set(CDConvex* c1, CDConvex* c2){
		convex[0] = c1;
		convex[1] = c2;
	}
	CDConvexPair():lastContactCount(-2){}
};
///	ʌ`yȂSgݍ킹
class CDConvexPairs:public UTCombination< UTRef<CDConvexPair> >{
public:
};
class CDFramePair;
///	WIg̃yA
class CDGeometryPair:public UTRefCount{
public:
	UTRef<CDGeometry> geometry[2];	///<	WIg
	UTRef<SGFrame> frame[2];		///<	WIgt[
	Affinef posture[2];				///<	t[̊t[̎p([NGA)
	CDConvexPairs convexPairs;		///<	2̃WIg\ʌ`ׂ̂ẴyA
	void Set(CDFramePair* fp, CDGeometry* g1, SGFrame* f1, CDGeometry* g2, SGFrame* f2);
	virtual ~CDGeometryPair(){}
};
///	WIg̃yȂSgݍ킹
class CDGeometryPairs:public UTCombination< UTRef<CDGeometryPair> >{
public:
};

///	Փ˔茋
class CDIntersection{
public:
	CDConvexPair* convexPair;
	CDGeometryPair* geometryPair;
	CDIntersection(CDConvexPair* cp, CDGeometryPair* gp): convexPair(cp), geometryPair(gp){}
};
///	Փ˔茋ʂ̔z
	class CDIntersections:public std::vector<CDIntersection>{
};

	
///	[UR[h̊{NX
typedef SGObject CDUserRecord;

///	[UR[h̃Rei^
class CDUserRecords:public std::vector< UTRef<CDUserRecord> >{
public:
};
///	Փ˔Ώۃt[D	t[ƃ[UR[h
class CDFrame:public UTRefCount{
public:
	CDFrame(){}
	CDFrame(SGFrame* f): frame(f){}
	CDUserRecords records;
	UTRef<SGFrame> frame;
};
/**	Փ˔Ώۂ̃t[̃yA.
t[͓KwłȂ΂ȂȂ(e₻Posture͍lȂ)D

	t[̏L֌W
	SGFrame
	 +-qSGFrame
	 +-CDMesh
		    +-CDPolyhedron
	t[(SGFrame)́C̃WIg(CDGeometries)Ǝqt[D
	WIg(CDGeometry) ̒ɂ́C̃WIg(CDMesh)
	ƒPȓʌ`(CDPolyhedron)D
	CDMesh ͕ConvexD
*/
class CDFramePair{	
public:
	///	XiANeBuǂ̃tOD
	class CDIsActive: public std::vector<bool>{
		bool bOr;
	public:
		void Update(){
			bOr = false;
			for(unsigned i=0; i<size(); ++i){
				bOr |= at(i);
			}
		}
		bool GetOr(){ return bOr; }
	};
	struct GetConv{
		CDConvexPairs* operator()(CDGeometryPairs::iterator it){ return *it ? &(*it)->convexPairs : NULL; }
	};
	///	ʌ`̃yA񋓂邽߂̃C^[^
	struct CDConvexPairIt: UTComposedIterator<CDConvexPair*, CDGeometryPairs::iterator, CDConvexPairs::iterator, GetConv>{
		friend class CDCollisionEngine;
		CDConvexPairIt(){}
		CDConvexPairIt(CDGeometryPairs::iterator b, CDGeometryPairs::iterator e, CDConvexPairs::iterator c){
			Init(b, e, c);
		}
	};

	///	ŌɐڐG
	unsigned lastContactCount;
	///	ANeBuǂ̃tODCDCollisionEnginẽXi̐tOD
	CDIsActive isActive;
	///	Ώۃt[
	UTRef<CDFrame> frame[2];
	///	Փ˂yAׂz
	CDIntersections intersections;
	///	WIg̃yȂSgݍ킹
	CDGeometryPairs geometryPairs;	

	///
	CDFramePair();
	///	t[̃Zbg
	void Set(CDFrame* f1, CDFrame* f2);
	///	ŜANeBuǂ
	bool IsActive(){ return isActive.GetOr(); }
	///	Xi i ANeBuǂ
	bool IsActive(int i){ return isActive[i]; }
	///	NA
	void Clear(){
		frame[0] = frame[1] = NULL;
		geometryPairs.clear();
	}
	///	L_T
	bool Detect(int count);
	///	ڐGĂ邩ǂ
	bool IsContact(SGScene* scene);

	///	ʌ`΂̗
	CDConvexPairIt ConvexPairBegin(){
		CDConvexPairIt rv;
		rv.InitAsBegin(geometryPairs.begin(), geometryPairs.end());
		return rv;
	}
	CDConvexPairIt ConvexPairEnd(){
		CDConvexPairIt rv;
		rv.InitAsEnd(geometryPairs.begin(), geometryPairs.end());
		return rv;
	}
protected:

	///	ڐG̃R[obN
	virtual void Found(CDConvexPair& cp, CDGeometryPair& gp){
		intersections.push_back(CDIntersection(&cp, &gp));
	}
	/**	CDConvexPair̔hNXg߂̎dg݁D
		̃NXpāC̊֐I[o[Ch邱ƂŁC
		CDConvexPaiȓɔhNXgpłD	*/
	virtual CDConvexPair* CreateConvexPair(){
		return new CDConvexPair;
	}
	/**	CDGeometryPair̔hNXg߂̎dg݁D
		̃NXpāC̊֐I[o[Ch邱ƂŁC
		CDGeometryPaiȓɔhNXgpłD	*/
	virtual CDGeometryPair* CreateGeometryPair(){
		return new CDGeometryPair;
	}
private:
	///	t[WIg񋓁D
	void EnumGeometry(SGFrames& frames, CDGeometries& geoms, SGFrame* f);
	friend class CDGeometryPair;
};
///	@ConvexPair
class CDConvexPairWithNormal: public CDConvexPair{
public:
	Vec3d normal;				///<	Փ˂̖@(01) (Global)
	Vec3d iNormal;				///<	ϕɂ@
	Vec3d center;				///<	2̍ŐN_̒Ԃ̓_CCDContactAnalysis::CalcNormal XVD
	double depth;				///<	Փ˂̐[FŋߖT_߂邽߂ɁC2̂𓮂D
};
///	@FramePair
class CDFramePairWithNormal :public CDFramePair{
protected:
	virtual CDConvexPair* CreateConvexPair(){
		return new CDConvexPairWithNormal;
	}
};

///	[UR[hƖ@ConvexPair
class CDConvexPairWithRecord: public CDConvexPairWithNormal{
public:
	CDUserRecords records;
};
///	CDConvexPairWithRecord̂߂CDFramePair R[hD
class CDFramePairWithRecord :public CDFramePairWithNormal, public UTRefCount{
public:
	CDUserRecords records;
protected:
	virtual CDConvexPair* CreateConvexPair(){
		return new CDConvexPairWithRecord;
	}
};
///	Փ˔̑Ώۂ̃t[΁D
class CDFramePairWithRecords: public UTCombination< UTRef<CDFramePairWithRecord> >{
public:
};

}
#endif
